{-# LANGUAGE CPP #-}
{-
parmap and mapReduce skeletons
-}

module ParMap where

import System.IO.Unsafe
import Control.Monad
import Control.Concurrent
import Data.List

import Control.Parallel.Eden

------------------------------------
-- Eden's parallel map
parMapEden,parMapFarm, parMapChunk, parMapWP
    :: (Trans a, Trans b) => (a -> b) -> [a] -> [b]
parMapEden = spawn . repeat . process 

--------------------------------------
-- Eden process farm (np processes, each working on a whole sublist)
farm :: (Trans a, Trans b) =>
        Int -> 
       (Int -> [a] -> [[a]]) -> -- n, distribute
       ([[b]] -> [b]) ->        -- combine
       (a -> b)   ->            -- worker process
       [a] -> [b]               -- input/output
farm np distr combine f inputs = combine results
    where results = spawnF (repeat (map f)) parts
          parts   = distr np inputs
--        combine (spawnF (repeat f) (distr np inputs))

parMapFarm f xs = farm noPe unshuffle shuffle f xs
          -- noPe :: Int -- number of nodes in parallel system

unshuffle :: Int -> [a] -> [[a]]
unshuffle n xs = [takeEach n (drop i xs) | i <- [0..n-1]]
 where takeEach n [] = []
       takeEach n (x:xs) = x : takeEach n (drop (n-1) xs)
-- inverse to unshuffle
shuffle :: [[a]] -> [a]
shuffle = concat . transpose

--------------------------------------
parMapChunk f xs = concat results
    where results = spawnF (repeat (map f)) parts
          parts   = splitIntoN noPe xs

-- dividing a list into n contiguous pieces (needs list length)
splitIntoN :: Int -> [a] -> [[a]]
splitIntoN n xs = takeIter parts xs
  where l = length xs
        parts = zipWith (+) ((replicate (l `mod` n) 1) ++ repeat 0)
                            (replicate n (l `div` n))
takeIter :: [Int] -> [a] -> [[a]]
takeIter [] [] = []
takeIter [] _ = error "elements left over"
takeIter (t:ts) xs = hs : takeIter ts rest
    where (hs,rest) = splitAt t xs

---------------------------------------
parMapWP = edenWP (max 1 (noPe-1)) 30 -- hard-coded prefetch here

edenWP :: (Trans t, Trans r) =>
            Int ->			-- #workers
            Int ->			-- prefetch
            (t -> r) ->                 -- worker
            [t] -> [r]
edenWP np prefetch f tasks = results
  where fromWorkers        = spawn workerProcs toWorkers
        workerProcs        = [process (zip [n,n..] . map f) | n<-[1..np]]
        toWorkers          = distribute tasks requests
        (newReqs, results) = (unzip . merge) fromWorkers
        requests           = initialReqs ++ newReqs
        initialReqs        = concat (replicate prefetch [1..np])
        distribute :: [t] -> [Int] -> [[t]]
        distribute tasks reqs = [taskList reqs tasks n | n<-[1..np]] 
          where taskList (r:rs) (t:ts) pe | pe == r    = t:(taskList rs ts pe)
                                          | otherwise  =    taskList rs ts pe
                taskList _      _      _  = []

---------------------------------------

parmapReduceStream :: (Trans a, Trans b) =>
                      Int -> -- no. of processes
                     (a -> b) -> -- mapped on input
                     (b -> b -> b) -> -- reduction (assumed commutative)
                     b -> -- neutral element for reduction
                     [a] -> b
parmapReduceStream np mapF redF neutral list
    = let sublists = unshuffle np list
          subFold = foldl' redF neutral . (map mapF)
          subRs = spawnF (repeat subFold) sublists
      in foldl redF neutral subRs

